use crate::api_types::{
    self,
    Color::{self, White},
    Coordinates, Move, PieceKind, PollResponse,
};
use std::{
    cell::RefCell,
    collections::{HashMap, HashSet},
    ops,
};

#[derive(Clone, Eq, PartialEq, Hash, Debug)]
pub(crate) struct Position {
    file: u8,
    rank: u8,
}

impl Position {
    fn from_coordinates(coord: Coordinates) -> Self {
        let file = coord.file.to_number();
        let rank = coord.rank.to_number();
        Position { file, rank }
    }
}

impl ops::Add<(i8, i8)> for &Position {
    type Output = Option<Position>;

    fn add(self, rhs: (i8, i8)) -> Self::Output {
        let file = self.file as i8 + rhs.0;
        let rank = self.rank as i8 + rhs.1;
        if !(0..=7).contains(&file) || !(0..=7).contains(&rank) {
            None
        } else {
            Some(Position {
                file: file as u8,
                rank: rank as u8,
            })
        }
    }
}

#[derive(Clone, Debug)]
pub(crate) struct Piece {
    kind: PieceKind,
    color: Color,
    visible: RefCell<HashSet<Position>>,
}

impl Piece {
    fn get_visible_in_direction(
        direction: (i8, i8),
        pos: &Position,
        state: &GameState,
    ) -> HashSet<Position> {
        let mut visible = HashSet::new();
        let mut current_square = pos.clone();

        loop {
            // keep moving in the specified direction, adding empty squares, until we reach edge of board
            // if we encounter a piece, that is the last visible square
            current_square = match &current_square + direction {
                None => break,
                Some(pos) => pos,
            };

            // add current square to visible
            visible.insert(current_square.clone());

            // path is blocked by piece, so cannot see/move past it
            if state.board.get(&current_square).is_some() {
                break;
            }
        }
        visible
    }

    fn generate_visible(&self, pos: &Position, state: &GameState) {
        let mut visible = HashSet::new();
        let mut move_directions: Vec<(i8, i8)> = Vec::new();

        match self.kind {
            PieceKind::Pawn => {
                match self.color {
                    Color::Black => {
                        // can capture diagonally one rank ahead
                        visible.extend(pos + (-1, -1));
                        visible.extend(pos + (1, -1));
                    }
                    Color::White => {
                        visible.extend(pos + (-1, 1));
                        visible.extend(pos + (1, 1));
                    }
                }
            }
            PieceKind::Rook => {
                move_directions.extend([(0, 1), (1, 0), (-1, 0), (1, 0)]);
            }
            PieceKind::Knight => {
                let knight_moves: Vec<(i8, i8)> = vec![
                    (-1, -2),
                    (-1, 2),
                    (1, -2),
                    (1, 2),
                    (-2, -1),
                    (-2, 1),
                    (2, -1),
                    (2, 1),
                ];
                visible.extend(knight_moves.iter().filter_map(|&mv| pos + mv));
            }
            PieceKind::Bishop => {
                move_directions.extend([(1, 1), (-1, -1), (-1, 1), (1, -1)]);
            }
            PieceKind::Queen => {
                move_directions.extend([
                    (0, 1),
                    (1, 0),
                    (-1, 0),
                    (1, 0),
                    (1, 1),
                    (-1, -1),
                    (-1, 1),
                    (1, -1),
                ]);
            }
            PieceKind::King => {
                let king_moves: Vec<(i8, i8)> = vec![
                    (1, 0),
                    (-1, 0),
                    (1, 1),
                    (-1, 1),
                    (1, -1),
                    (-1, -1),
                    (0, -1),
                    (0, 1),
                ];
                visible.extend(king_moves.iter().filter_map(|&mv| pos + mv));
            }
        }

        for direction in move_directions.drain(..) {
            visible.extend(Piece::get_visible_in_direction(direction, pos, state))
        }

        *self.visible.borrow_mut() = visible;
    }

    fn get_possible_moves(&self, piece_pos: &Position, state: &GameState) -> HashSet<Position> {
        let mut possible_moves = HashSet::new();

        // for all pieces except pawn, visible squares are squares you can move to
        // for pawns, visible squares are diagonally ahead, possible moves are either 1 or 2 squares
        // directly ahead
        if self.kind == PieceKind::Pawn {
            let (opposite_color, starting_rank, one_square, two_square) = match self.color {
                Color::Black => (Color::White, 6, (0, -1), (0, -2)),
                Color::White => (Color::Black, 1, (0, 1), (0, 2)),
            };

            // can move to visible squares diagonally ahead of a piece of opposite color is there
            // OR if it is the en passant target square
            for pos in self.visible.borrow().iter() {
                if let Some(piece) = state.board.get(pos) {
                    if piece.color == opposite_color {
                        possible_moves.insert(pos.clone());
                    }
                }
                if state.en_passant_target.as_ref() == Some(pos) {
                    possible_moves.insert(pos.clone());
                }
            }

            // can move to square directly ahead if no piece is there
            let one_square = (piece_pos + one_square).unwrap();
            if state.board.get(&one_square).is_none() {
                possible_moves.insert(one_square);
                // can move two squares ahead if we are on starting rank
                if piece_pos.rank == starting_rank {
                    let two_square = (piece_pos + two_square).unwrap();
                    if state.board.get(&two_square).is_none() {
                        possible_moves.insert(two_square);
                    }
                }
            }
        } else {
            // for all other pieces, can move to any visible square unless it contains
            // a piece of the same color (can capture opponent piece, but not own)
            possible_moves = self
                .visible
                .borrow()
                .iter()
                .filter(|&pos| {
                    if let Some(piece) = state.board.get(pos) {
                        // space is occupied, check if it is same color or not
                        piece.color != self.color
                    } else {
                        // space is empty, so able to move there
                        true
                    }
                })
                .cloned()
                .collect();

            // check if king can castle
            if self.kind == PieceKind::King {
                // king side castle - must still be allowed and have two empty squares between
                // king and rook, not visible to opponent
                if state.castle.get(&self.color).unwrap().can_castle_king_side
                    && state.board.get(&(piece_pos + (1, 0)).unwrap()).is_none()
                    && state.board.get(&(piece_pos + (2, 0)).unwrap()).is_none()
                    && !state
                        .visible
                        .get(&self.color.opposite())
                        .unwrap()
                        .contains(&(piece_pos + (1, 0)).unwrap())
                    && !state.player_in_check(&self.color)
                {
                    // move is represented as king moving two squares
                    possible_moves.insert((piece_pos + (2, 0)).unwrap());
                }
                // queen side castle - must still be allowed and have three empty squares between
                // king and rook
                if state.castle.get(&self.color).unwrap().can_castle_queen_side
                    && state.board.get(&(piece_pos + (-1, 0)).unwrap()).is_none()
                    && state.board.get(&(piece_pos + (-2, 0)).unwrap()).is_none()
                    && state.board.get(&(piece_pos + (-3, 0)).unwrap()).is_none()
                    && !state
                        .visible
                        .get(&self.color.opposite())
                        .unwrap()
                        .contains(&(piece_pos + (-1, 0)).unwrap())
                    && !state.player_in_check(&self.color)
                {
                    // move is represented as king moving two squares
                    possible_moves.insert((piece_pos + (-2, 0)).unwrap());
                }
            }
        }

        // final validation - possible moves must not lead to check
        let possible_moves = possible_moves
            .iter()
            .filter(|&to_pos| {
                let mut new_state = state.clone();
                // apply move
                new_state.apply_move(piece_pos.clone(), to_pos.clone());
                //new_state.generate_visible();
                !new_state.player_in_check(&self.color)
            })
            .cloned()
            .collect();
        possible_moves
    }
}

#[derive(Clone, Debug)]
struct CastleState {
    can_castle_king_side: bool,
    can_castle_queen_side: bool,
}

#[derive(Clone, Debug, Default)]
pub(crate) struct GameState {
    board: HashMap<Position, Piece>,
    visible: HashMap<Color, HashSet<Position>>,
    en_passant_target: Option<Position>,
    castle: HashMap<Color, CastleState>,
    allowed_turn: Color,
}

impl GameState {
    pub(crate) fn starting_positions() -> Self {
        let mut this = Self::default();

        for (file, rank, kind, color) in INITIAL_PIECES {
            this.board.insert(
                Position { file, rank },
                Piece {
                    kind,
                    color,
                    visible: RefCell::new(HashSet::new()),
                },
            );
        }

        this.generate_visible();

        this.castle.insert(
            Color::Black,
            CastleState {
                can_castle_king_side: true,
                can_castle_queen_side: true,
            },
        );
        this.castle.insert(
            Color::White,
            CastleState {
                can_castle_king_side: true,
                can_castle_queen_side: true,
            },
        );

        this
    }

    pub(crate) fn to_serializable(&self) -> PollResponse {
        let board_state = self
            .board
            .iter()
            .map(|(pos, piece)| {
                (
                    api_types::File::from_number(pos.file).unwrap(),
                    api_types::Rank::from_number(pos.rank).unwrap(),
                    piece.kind.clone(),
                    piece.color.clone(),
                )
            })
            .collect();

        let game_state = api_types::GameState::from_color(self.allowed_turn.clone());

        PollResponse {
            board_state,
            game_state,
        }
    }

    fn generate_visible(&mut self) {
        let mut black_visible = HashSet::new();
        let mut white_visible = HashSet::new();

        for (pos, piece) in self.board.iter() {
            piece.generate_visible(pos, self);
            match piece.color {
                Color::Black => {
                    black_visible.extend(piece.visible.borrow().clone());
                }
                White => {
                    white_visible.extend(piece.visible.borrow().clone());
                }
            }
        }

        let mut visible = HashMap::new();
        visible.insert(Color::Black, black_visible);
        visible.insert(Color::White, white_visible);

        self.visible = visible;
    }

    // determines if given color king is in check
    fn player_in_check(&self, color: &Color) -> bool {
        let opponent_visible = self.visible.get(&color.opposite()).unwrap();
        let king_color_collection: Vec<(&Position, &Piece)> = self
            .board
            .iter()
            .filter(|(_, piece)| piece.kind == PieceKind::King && &piece.color == color)
            .collect();
        let (king_pos, _king) = &king_color_collection[0];
        opponent_visible.contains(king_pos)
    }

    fn apply_move(&mut self, from: Position, to: Position) {
        let piece = self.board.remove(&from).unwrap();

        // if pawn is moving to en passant square, remove captured pawn
        if piece.kind == PieceKind::Pawn && Some(&to) == self.en_passant_target.as_ref() {
            let pawn_pos = &to
                + match piece.color {
                    Color::Black => (0, 1),
                    White => (0, -1),
                };
            self.board.remove(&pawn_pos.unwrap());
        }
        // update en passant square if a pawn moves 2 squares ahead
        if piece.kind == PieceKind::Pawn
            && ((piece.color == Color::White && from.rank == 1 && to.rank == 3)
                || (piece.color == Color::Black && from.rank == 6 && to.rank == 4))
        {
            // en passant square is square behind where pawn moves to
            self.en_passant_target = &to
                + match piece.color {
                    Color::Black => (0, 1),
                    White => (0, -1),
                };
        } else {
            self.en_passant_target = None;
        }

        // check if we are castling
        if piece.kind == PieceKind::King {
            if self.castle.get(&piece.color).unwrap().can_castle_king_side
                && Some(&to) == (&from + (2, 0)).as_ref()
            {
                // perform king side castle by moving rook (king gets moved to to_pos)
                let rook_pos = Position {
                    file: 7,
                    rank: to.rank,
                };
                if let Some(rook) = self.board.remove(&rook_pos) {
                    self.board.insert((&rook_pos + (-2, 0)).unwrap(), rook);
                }
            } else if self.castle.get(&piece.color).unwrap().can_castle_queen_side
                && Some(&to) == (&from + (-2, 0)).as_ref()
            {
                // perform queen side castle
                let rook_pos = Position {
                    file: 0,
                    rank: to.rank,
                };
                if let Some(rook) = self.board.remove(&rook_pos) {
                    self.board.insert((&rook_pos + (3, 0)).unwrap(), rook);
                }
            }
            // even if not castling, if you move king you can no longer castle
            self.castle
                .get_mut(&piece.color)
                .unwrap()
                .can_castle_queen_side = false;
            self.castle
                .get_mut(&piece.color)
                .unwrap()
                .can_castle_king_side = false;
        }

        // if moving rook from starting position, can no longer castle on that side
        if piece.kind == PieceKind::Rook {
            if from.file == 7 {
                self.castle
                    .get_mut(&piece.color)
                    .unwrap()
                    .can_castle_king_side = false;
            } else if from.file == 0 {
                self.castle
                    .get_mut(&piece.color)
                    .unwrap()
                    .can_castle_queen_side = false;
            }
        }

        // move the piece
        self.board.insert(to, piece);

        self.generate_visible()
    }

    pub(crate) fn handle(&mut self, action: Move) {
        let from_pos = Position::from_coordinates(action.from);
        let to_pos = Position::from_coordinates(action.to);

        let piece = match self.board.get(&from_pos) {
            None => {
                panic!("No piece found at position specified in move");
            }
            Some(piece) => piece,
        }
        .clone();

        if piece.get_possible_moves(&from_pos, self).contains(&to_pos) {
            self.apply_move(from_pos, to_pos);
            self.allowed_turn = self.allowed_turn.opposite();
        }
    }
}

const INITIAL_PIECES: [(u8, u8, PieceKind, Color); 32] = [
    (0, 0, PieceKind::Rook, Color::White),
    (1, 0, PieceKind::Knight, Color::White),
    (2, 0, PieceKind::Bishop, Color::White),
    (3, 0, PieceKind::Queen, Color::White),
    (4, 0, PieceKind::King, Color::White),
    (5, 0, PieceKind::Bishop, Color::White),
    (6, 0, PieceKind::Knight, Color::White),
    (7, 0, PieceKind::Rook, Color::White),
    (0, 1, PieceKind::Pawn, Color::White),
    (1, 1, PieceKind::Pawn, Color::White),
    (2, 1, PieceKind::Pawn, Color::White),
    (3, 1, PieceKind::Pawn, Color::White),
    (4, 1, PieceKind::Pawn, Color::White),
    (5, 1, PieceKind::Pawn, Color::White),
    (6, 1, PieceKind::Pawn, Color::White),
    (7, 1, PieceKind::Pawn, Color::White),
    (0, 6, PieceKind::Pawn, Color::Black),
    (1, 6, PieceKind::Pawn, Color::Black),
    (2, 6, PieceKind::Pawn, Color::Black),
    (3, 6, PieceKind::Pawn, Color::Black),
    (4, 6, PieceKind::Pawn, Color::Black),
    (5, 6, PieceKind::Pawn, Color::Black),
    (6, 6, PieceKind::Pawn, Color::Black),
    (7, 6, PieceKind::Pawn, Color::Black),
    (0, 7, PieceKind::Rook, Color::Black),
    (1, 7, PieceKind::Knight, Color::Black),
    (2, 7, PieceKind::Bishop, Color::Black),
    (3, 7, PieceKind::Queen, Color::Black),
    (4, 7, PieceKind::King, Color::Black),
    (5, 7, PieceKind::Bishop, Color::Black),
    (6, 7, PieceKind::Knight, Color::Black),
    (7, 7, PieceKind::Rook, Color::Black),
];
